﻿#region Header

/*
Behavioral Rating of Dancing Human Crowds based on Motion Patterns
By

Pascal Hauser 
Dipl. Ing. in Informatik, Hochschule für Technik Rapperswil, 2006
Master Thesis, Hochschule für Technik Rapperswil, 2008-2010

and

Raphael Gfeller
Dipl. Ing. in Informatik, Hochschule für Technik Rapperswil, 2006
Master Thesis, Hochschule für Technik Rapperswil, 2008-2010

*/

#endregion

#region Usings

using System;
using System.Collections.Generic;
using paravili.Services;
using Sebarf.Services.Interfaces;

#endregion

namespace paravili.Steps {
	/// <summary>
	/// tries to find movement between the given list of points
	/// </summary>
	public class FindMovementBetweenClustersByUsingServiceByListStructure :
		ProcessStepConverterWithMeasurement<IList<IList<ColoredPoint>>, IList<Cluster>> {
		#region Public Properties

		[ConfigurableNumericValue(Name = "MaxDistanceBetweenTwoCluster", RangeFrom = 10, RangeTo = 500)]
		public int MaxDistanceBetweenTwoCluster { get; set; }

		[ConfigurableNumericValue(Name = "MaxFramesBetweenTwoCluster", RangeFrom = 0, RangeTo = 50)]
		public int MaxFramesBetweenTwoCluster { get; set; }

		[ConfigurableNumericValue(Name = "WeightOfDistance", RangeFrom = 1, RangeTo = 10)]
		public int WeightOfDistance { get; set; }

		[ConfigurableNumericValue(Name = "WeightOfColour", RangeFrom = 1, RangeTo = 10)]
		public int WeightOfColour { get; set; }

		[ConfigurableNumericValue(Name = "WeightConvexHullSize", RangeFrom = 1, RangeTo = 10)]
		public int WeightConvexHullSize { get; set; }

		[ConfigurableNumericValue(Name = "MaxLinkedClusters", RangeFrom = 1, RangeTo = 50)]
		public int MaxLinkedClusters { get; set; }

		[ServiceRequest]
		public ITimeService TimeService { get; set; }

		[ServiceRequest]
		public IResetHandlerService ResetHandlerService {
			set { value.Reset += OnReset; }
		}

		[ServiceRequest]
		public IDoMatchingService2 DoMatchingService { set; get; }

		#endregion

		#region Public Methods

		public FindMovementBetweenClustersByUsingServiceByListStructure() {
			MaxDistanceBetweenTwoCluster = 30;
			MaxFramesBetweenTwoCluster = 8;
			WeightOfDistance = 2;
			WeightOfColour = 1;
			WeightConvexHullSize = 1;
			MaxLinkedClusters = 5;
			DoReset();
		}

		#endregion

		#region Private Methods

		protected override IList<Cluster> Convert(IList<IList<ColoredPoint>> toProcess) {
			if (toProcess != null) {
				// prepare data
				foreach (var pointList in toProcess) {
					m_availableClusters.Add(new Cluster(pointList, TimeService.GetTime()));
				}
			}
			IList<Cluster> toReturn = FindMovement();
			return toReturn;
		}

		/// <summary>
		/// Reset handler
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args"></param>
		private void OnReset(Object sender, EventArgs args) {
			DoReset();
		}

		/// <summary>
		/// resets the step
		/// </summary>
		private void DoReset() {
			m_availableClusters.Clear();
			m_minDistance = double.MaxValue;
			m_minColor = double.MaxValue;
			m_minSize = double.MaxValue;
			m_maxColor = double.MinValue;
			m_maxDistance = double.MinValue;
			m_maxSize = double.MinValue;
		}

		/// <summary>
		/// Calculates how close two given clusters are
		/// </summary>
		/// <param name="first"></param>
		/// <param name="last"></param>
		/// <returns></returns>
		private double MatchFunction(Cluster first, Cluster last) {
			if (first == last) {
				return double.MinValue;
			}
			if (last.PreviousCluster != null) {
				return double.MinValue;
			}
			double matchingPoints = 0;
			double distance =
				Math.Pow(
					Math.Pow(first.AverageXValue - last.AverageXValue, 2) + Math.Pow(first.AverageYValue - last.AverageYValue, 2),
					0.5);
			if (MaxDistanceBetweenTwoCluster < distance) {
				return double.MinValue;
			}
			m_minDistance = Math.Min(m_minDistance, distance);
			m_maxDistance = Math.Max(m_maxDistance, distance);

			matchingPoints -= Normalize(distance, m_minDistance, m_maxDistance, WeightOfDistance);
			// (distance / ReturnOneIfNull(m_maxDistance - m_minDistance) * WeightOfDistance);

			int timeDiff = Math.Abs(first.Time - last.Time);

			if (timeDiff == 0) {
				return double.MinValue;
			}
			if (first.Time != TimeService.GetTime()) {
				return double.MinValue;
			}

			double colorDifferenceDifference =
				Math.Pow(
					Math.Pow(first.ColorY - last.ColorY, 2) + Math.Pow(first.ColorX - last.ColorX, 2) +
					Math.Pow(first.ColorZ - last.ColorZ, 2), 0.5);

			m_minColor = Math.Min(m_minColor, colorDifferenceDifference);
			m_maxColor = Math.Max(m_maxColor, colorDifferenceDifference);
			matchingPoints -= Normalize(colorDifferenceDifference, m_minColor, m_maxColor, WeightOfColour);

			double diffConvexHullSize = Math.Abs(first.HulPerimeterSize - last.HulPerimeterSize);

			m_minSize = Math.Min(m_minSize, diffConvexHullSize);
			m_maxSize = Math.Max(m_maxSize, diffConvexHullSize);

			matchingPoints -= Normalize(diffConvexHullSize, m_minSize, m_maxSize, WeightConvexHullSize);

			return matchingPoints;
		}

		/// <summary>
		/// helper function to avoid null values
		/// </summary>
		/// <param name="d"></param>
		/// <returns></returns>
		private static double ReturnOneIfNull(double d) {
			//if (d == 0) {
			//		return 1;
			//	}
			return d;
		}

		private static double Normalize(double value, double minValue, double maxValue, double weight) {
			if (value == minValue && minValue == maxValue) {
				return 0;
			}
			double ratio = 100.0 / Math.Abs(ReturnOneIfNull(maxValue - minValue)) * (value - ReturnOneIfNull(minValue));
			if (ratio == 0) {
				return 0;
			}
			return Math.Pow(Math.E, ratio / 10.0 * weight);
		}

		/// <summary>
		/// match clusters together
		/// </summary>
		/// <returns></returns>
		private IList<Cluster> FindMovement() {
			if (m_availableClusters.Count < 1) {
				return null;
			}
			// remove to old clusters
			int maxTime = MaxFramesBetweenTwoCluster;
			int time = TimeService.GetTime();
			m_availableClusters.RemoveAll(c => time - c.Time > maxTime);

			// rember found clusters
			IList<Cluster> toReturn = new List<Cluster>();
			IList<Cluster> toRemove = new List<Cluster>();

			Cluster lastCluster;
			Cluster[] clusters = m_availableClusters.ToArray();

			// create input data for the matching service
			var inputData = new MatchingServiceInputData();
			int overallStartIndex = clusters.Length + 1;
			int overallEndIndex = clusters.Length + 2;
			inputData.StartIndex = overallStartIndex;
			inputData.TargetIndex = overallEndIndex;
			var arcs = new List<MatchingServiceInputData.Arc>();
			for (int i = 0; i < clusters.Length; i++) {
				for (int j = 0; j < clusters.Length; j++) {
					double cost = MatchFunction(clusters[i], clusters[j]);
					if (cost != double.MinValue) {
						arcs.Add(new MatchingServiceInputData.Arc { Capacity = 1, Cost = (int)cost * -1, FromIndex = i, ToIndex = j });
						arcs.Add(new MatchingServiceInputData.Arc { Capacity = 1, Cost = 0, FromIndex = overallStartIndex, ToIndex = i });
						arcs.Add(new MatchingServiceInputData.Arc { Capacity = 1, Cost = 0, FromIndex = j, ToIndex = overallEndIndex });
					}
				}
			}
			inputData.Arcs = arcs.ToArray();
			Match[] matches = DoMatchingService.FindMatch(inputData);
			foreach (Match match in matches) {
				int firstClusterIndex = match.FromIndex;
				int lastClusterIndex = match.ToIndex;
				lastCluster = clusters[lastClusterIndex];
				Cluster firstCluster = clusters[firstClusterIndex];
				lastCluster.PreviousCluster = firstCluster;
				toReturn.Add(lastCluster);
				toRemove.Add(firstCluster);
			}
			foreach (Cluster cluster in toRemove) {
				m_availableClusters.Remove(cluster);
			}
			StatisticValuesProviderService.SetStatisticValue("Movements.Found", toReturn.Count);
			return toReturn;
		}

		#endregion

		#region Private Fields

		private readonly List<Cluster> m_availableClusters = new List<Cluster>();
		private double m_maxColor;
		private double m_maxDistance;
		private double m_maxSize;
		private double m_minColor;
		private double m_minDistance;
		private double m_minSize;

		#endregion
	}
}