package lpl;

/**
 * Wrap all procedures and functions exported by the corresponding Windows
 * Dynamic Linked Library (lplj.dll). It uses the Java Native Interface
 * 
 * @author <a href="mailto:matthias.buchs@unifr.ch">Matthias Buchs </a>
 * @author <a href="mailto:alexandre.gachet@unifr.ch">Alexandre Gachet</a>
 * @author <a href="mailto:tony.huerlimann@unifr.ch">Tony Huerlimann</a>
 */

public class NativeLPL {
    /**
     * The constructor is made private to enforce the static functionality of this class
     */
    private NativeLPL() {
        ;
    }

    /**
     * Wrap the procedure LPLcompile(lplC:TProc,logC:TProcC); stdcall; export;
     * local version. Before compiling this method sets the desired callbacks by
     * calling LPLsetCallbacks. After compilation these callbacks are
     * immediately removed.
     * 
     * @param options a String object describing LPL switches
     * @param lplCallback an instance implementing the LPLCallback interface
     * @param logCallback an instance implementing the LogCallback interface
     * @return a LPL-specific status code
     */
    public static native int LPLcompileWithCallbacks(String options, LPLCallback lplCallback, LogCallback logCallback);

    /**
     * Wrap the procedure LPLinitParam; stdcall; export
     */
    public static native void LPLinitParam();

    /**
     * Wrap the procedure LPLinit(Fn:pChar; maxt,maxa,maxr:integer; cb:TProc);
     * stdcall; export; local version
     * 
     * @param fn a String object pointing to the model filename
     * @param maxt LPL parameter (see LPL documentation)
     * @param maxa LPL parameter (see LPL documentation)
     * @param maxr LPL parameter (see LPL documentation)
     */
    public static native void LPLinit(String fn, int maxt, int maxa, int maxr);

    /**
     * Wrap the function LPLcompile(opt:pChar):integer; stdcall; export;
     * 
     * @param s a String object describing LPL switches
     * @return a LPL-specific status code
     */
    public static native int LPLcompile(String s);

    /**
     * Wrap the procedure LPLfree; stdcall; export;
     */
    public static native void LPLfree();

    /**
     * Wrap the function LPLwhere; stdcall; export;
     * 
     * @return a LPL-specific code
     */
    public static native int LPLwhere();

    /**
     * Wrap the procedure LPLGetErrMessage(n:integer; var msg:PChar); stdcall;
     * export;
     * 
     * @param n error number
     * @return a String object describing the error
     */
    public static native String LPLgetErrMessage(int n);

    /**
     * Wrap the function GetErrorNr:integer; stdcall; export;
     * 
     * @return LPL error number
     */
    public static native int LPLgetError();

    /**
     * Wrap the procedure SetErrorNr(n:integer); stdcall; export;
     * 
     * @param n error number
     */
    public static native void LPLsetError(int n);

    /**
     * Wrap the procedure LPLgetP(Arg1: JInt); stdcall;
     * 
     * @param which information to retrieve (see LPL manual)
     * @return a String object representing the required information
     */
    public static native String LPLgetP(int which);

    /**
     * Wrap the procedure LPLsetP(Arg1: JInt; Arg2: JString); stdcall;
     * 
     * @param which information to overwrite (see LPL manual)
     * @param sP new information
     */
    public static native void LPLsetP(int which, String sP);

    /**
     * Wrap the procedure LPLsolve; stdcall; export;
     * 
     * @param s parameters passed to the LPL solver
     * @return the return code of the LPL solver
     */
    public static native int LPLsolve(String s);

    /**
     * Wrap the function LPLSaveSnapshot; stdcall; export;
     * 
     * @param sP1 the name of the snapshot
     */
    public static native void LPLsaveSnapshot(String sP1);

    /**
     * Wrap the procedure LPLLoadSnapshot(); stdcall; export;
     * 
     * @param sP1 the name of the snapshot
     * @param add guiding the snapshot loading (e.g. multiple snapshot)
     */
    public static native void LPLloadSnapshot(String sP1, int add);

    /**
     * Wrap the procedure LPLgenACCESSdb; stdcall; export;
     */
    public static native void LPLgenACCESSdb();

    /**
     * Wrap the procedure LPLgetHandle(String name); stdcall; export;
     * 
     * @param name the element name
     * @return an integer handle to the specified element
     */
    public static native int LPLgetHandle(String name);

    /**
     * Wrap the procedure LPLgetFocusHandle; stdcall; export;
     * 
     * @param h an LPLhandle as int to the focused entity.
     */
    public static native int LPLgetFocusHandle();

    /**
     * Wrap the procedure LPLsetFocus; stdcall; export;
     * 
     * @param h an LPLhandle as int to set the focus to
     */
    public static native void LPLsetFocus(int h);

    /**
     * Wrap the procedure LPLisFocus; stdcall; export;
     * 
     * @param name the qualified name of the entity focus the entity.
     * @returns true if last focus was successful
     */
    public static native boolean LPLisFocus();

    /**
     * Wrap the procedure LPLsetFirstEntity; stdcall; export; focus the first
     * entity of the genus.
     * 
     * @param genus the genus number
     */
    public static native void LPLsetFirstEntity(int genus);

    /**
     * Wrap the procedure LPLnextEntity; stdcall; export; focus the next entity
     * of the genus.
     * 
     * @param genus the genus number
     */
    public static native void LPLnextEntity(int genus);

    /**
     * Wrap the procedure LPLgetElementList; stdcall; export;
     * 
     * @param n the index of the desired element list
     * @param enam method will return the element names if equal to 0, the
     *            STRING aliases if equal to 1
     * @return returns the n-th element list of the entity name.
     */
    public static native String LPLgetElementList(int n, int enam);

    /**
     * Wrap the procedure LPLgetAttr; stdcall; export;
     * 
     * @param i the parameter of the information requested
     * @return information about the focused entity
     */
    public static native String LPLgetAttr(int i);


    /** Wrap the procedure LPLPivotSetParam; stdcall; export;
     * 
     * @param which the parameter of the information requested
     * @param n value of the parameter
     */
    public static native void LPLPivotSetParam(int which, int n);

    /** Wrap the procedure LPLPivotGetParam; stdcall; export;
     * 
     * @param which the parameter of the information requested
     */
    public static native int LPLPivotGetParam(int which);

    /** Wrap the procedure LPPivotInit; stdcall; export;
     * 
     */
    public static native void LPLPivotInit();

    /** Wrap the procedure LPPivotX; stdcall; export;
     *
     * @return x-coor in pivot table of current element 
     */
    public static native int LPLPivotX();

    /** Wrap the procedure LPLPivotY(); stdcall; export;
     *
     * @return y-coor in pivot table of current element
     */
    public static native int LPLPivotY();

    /** Wrap the procedure LPPivotData; stdcall; export;
     *
     * @return pivot cell content at (x,y) 
     */
    public static native String LPLPivotData();


    /** Wrap the procedure LPLsetSelect; stdcall; export;
     * 
     * @param n the n-th element (if 0 then all/none/inverse)
     * @param what : 1:select, 0:unselect, (-1:inverse all, if n=0)
     */
    public static native void LPLsetSelect(int n, int what);


    /** Wrap the procedure LPLgetValue; stdcall; export;
     * 
     * @param n the position of the entity name
     * @param attr the parameter of the information requested
     * @return the value of the focused entity at the position n
     */
    public static native double LPLgetValue(int n, int attr);

    /**
     * Wrap the procedure LPLgetValueS; stdcall; export;
     * 
     * @param n the position of the entity name
     * @param attr the parameter of the information requested
     * @return the string value of the focused entity at the position n
     */
    public static native String LPLgetValueS(int n, int attr);

    /**
     * Wrap the procedure LPLgetM; stdcall; export;
     * 
     * @param attr the parameter of the information requested
     * @return global information
     */
    public static native int LPLgetM(int attr);

    /**
     * Wrap the procedure LPLgetDependencyFrom(int n); stdcall; export;
     * 
     * @param n the position of the dependency
     * @return the source handle of the specified dependency
     */
    public static native int LPLgetDependencyFrom(int n);

    /**
     * Wrap the procedure LPLgetDependencyTo(int n); stdcall; export;
     * 
     * @param n the position of the entry
     * @return the target handle of the specified dependency
     */
    public static native int LPLgetDependencyTo(int n);

    /**
     * Wrap the procedure LPLgetDependencyKind(int n); stdcall; export;
     * 
     * @param n the position of the dependency
     * @return the type of the specified dependency
     */
    public static native char LPLgetDependencyKind(int n);

    static {
        System.loadLibrary("lplj");
    }
}